#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['rnnr']

package_data = \
{'': ['*']}

install_requires = \
['tqdm']

extras_require = \
{'dev': ['flake8', 'flake8-mypy', 'yapf'],
 'doc': ['Sphinx ~=1.8.5',
         'sphinx_rtd_theme',
         'sphinx-autodoc-typehints ~=1.6.0'],
 'test': ['pytest ~=4.3.0', 'pytest-cov']}

setup(name='rnnr',
      version='0.0.6',
      description='rnnr: neural network runner',
      author='Kemal Kurniawan',
      author_email='kemal@kkurniawan.com',
      url='https://github.com/kmkurn/rnnr',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6,<4',
     )
